"""Shared functionality between GPU and CPU correlators."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np
from scipy.ndimage import laplace as laplace_filter
from tqdm import tqdm

if TYPE_CHECKING:
    from pyopencl import Image
    from pyopencl.array import Array as ClArray

f32 = np.float32
i32 = np.int32

T = TypeVar("T", np.ndarray, "ClArray")
I = TypeVar("I", np.ndarray, "Image")  # noqa: E741


@dataclass
class Vars(Generic[T, I]):
    """Non-complex GPU arrays."""

    target: T
    template: I
    mask: I
    lcc_mask: T
    target2: T
    rot_template: T
    rot_mask: T
    rot_mask2: T
    gcc: T
    ave: T
    ave2: T
    lcc: T
    rot: T


@dataclass
class VarsFT(Generic[T]):
    """Fourier transformed (complex) arrays."""

    target: T
    target2: T
    template: T
    mask: T
    mask2: T
    ave: T
    ave2: T
    lcc: T
    gcc: T


def get_lcc_mask(target: np.ndarray) -> np.ndarray:
    """Compute the local cross correlation (LCC) mask.

    Note that the mask is equal to all target voxels where the values
    exceed 5% of the maximum voxel value. Only these voxels are used for
    computing the LCC in the `calc_lcc_and_take_best` kernel function.
    """
    return target > target.max() * 0.05


def normalize_template(template: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """Normalize the template structure cf. A.M. Roseman (2000)."""
    norm_template = template * mask
    # normalize template;
    ind = mask != 0
    norm_template[ind] -= norm_template[ind].mean()
    norm_template[ind] /= norm_template[ind].std()
    # multiply again for core-weighted correlation score
    return norm_template * mask


def get_ft_shape(target: np.ndarray) -> tuple:
    """Returns shape of fourier transformed target."""
    return target.shape[:-1] + (target.shape[-1] // 2 + 1,)


def get_normalization_factor(mask: np.ndarray) -> np.float32:
    """Precompute the normalization factor for use in the LCC computing kernel"""
    norm_factor = np.not_equal(mask, 0).sum(dtype=np.float32)
    if norm_factor == 0:
        raise ValueError("Zero-filled mask is not allowed.")
    return norm_factor


class Correlator(ABC):
    vars: Vars
    vars_ft: VarsFT
    rfftn: Callable
    irfftn: Callable
    conj_multiply: Callable
    square: Callable
    laplace: bool
    target: np.ndarray
    lcc: np.ndarray
    rot: np.ndarray

    @abstractmethod
    def __init__(self):
        """Initialize the correlator along with the above class properties."""
        pass

    @abstractmethod
    def _set_template_var(self, template: np.ndarray):
        """Set the Vars.template variable in-place."""
        pass

    @abstractmethod
    def _set_mask_var(self, mask: np.ndarray):
        """Set the Vars.mask variable in-place."""

    def set_template(self, template: np.ndarray, mask: np.ndarray):
        """Set the template structure that you want to fit in the target density.

        Can be used to try to fit a different template to the same target structure
        without recomputing the kernels.

        Args:
            template: the template structure that you want to fit in the target density,
                should have been regridded to the same grid as the target density.
        """
        if template.shape != self.target.shape:
            raise ValueError("Shape of template does not match the target.")

        if self.laplace:
            template = laplace_filter(template, mode="wrap")

        # Precompute the normalization factor for use in the LCC computing kernel
        self.norm_factor = get_normalization_factor(mask)

        template = normalize_template(template, mask)
        self._set_template_var(template)
        self._set_mask_var(mask)

        # Reset lcc and rot values after (re)setting the template
        self.lcc[:] = 0.0
        self.rot[:] = 0

    @abstractmethod
    def rotate_grids(self, rotmat: np.ndarray):
        """Rotate the template and mask using the rotational matrix."""
        pass

    def compute_gcc(self):
        """Compute the global cross-correlation.

        Ref doi:10.3934/biophy.2015.2.73. Equation 3."""
        self.rfftn(self.vars.rot_template, self.vars_ft.template)
        self.conj_multiply(self.vars_ft.template, self.vars_ft.target, self.vars_ft.gcc)
        self.irfftn(self.vars_ft.gcc, self.vars.gcc)

    def compute_sq_avg_density(self):
        """Compute the square of the average core-weighted density.

        Ref doi:10.3934/biophy.2015.2.73. Equation 4."""
        self.rfftn(self.vars.rot_mask, self.vars_ft.mask)
        self.conj_multiply(self.vars_ft.mask, self.vars_ft.target, self.vars_ft.ave)
        self.irfftn(self.vars_ft.ave, self.vars.ave)

    def compute_avg_sq_density(self):
        """Compute the average of the squared core-weighted density.

        Ref doi:10.3934/biophy.2015.2.73. Equation 5."""
        self.square(self.vars.rot_mask, self.vars.rot_mask2)
        self.rfftn(self.vars.rot_mask2, self.vars_ft.mask2)
        self.conj_multiply(self.vars_ft.mask2, self.vars_ft.target2, self.vars_ft.ave2)
        self.irfftn(self.vars_ft.ave2, self.vars.ave2)

    @abstractmethod
    def compute_lcc_score_and_take_best(self, n: int):
        """Compute the LCC score and store best result.

        Args:
            n: iteration number.
        """
        pass

    def compute_rotation(self, n: int, rotmat: np.ndarray):
        """Compute a single rotation.

        Args:
            n: rotation number.
            rotmat: rotation matrix for this rotation.
        """
        self.rotate_grids(rotmat)
        self.compute_gcc()
        self.compute_sq_avg_density()
        self.compute_avg_sq_density()
        self.compute_lcc_score_and_take_best(n)

    @abstractmethod
    def scan(self, progress: partial[tqdm] | None):
        pass



from unittest import TestCase, main

import numpy as np

from powerfit_em.shape_descriptor import ShapeDescriptor


class TestShapeDescriptor(TestCase):
    def test_relative_shape_anisotropy(self):
        # coordinates on a line
        coor = np.zeros((3, 10), dtype=np.float64)
        coor[0] = np.arange(10)
        sd = ShapeDescriptor(coor)
        self.assertAlmostEqual(sd.relative_shape_anisotropy, 1)

        # points on a unit sphere
        coor = np.zeros((3, 6), dtype=np.float64)
        coor[0] = [1, -1, 0, 0, 0, 0]
        coor[1] = [0, 0, 1, -1, 0, 0]
        coor[2] = [0, 0, 0, 0, 1, -1]
        sd = ShapeDescriptor(coor)
        self.assertAlmostEqual(sd.relative_shape_anisotropy, 0)


if __name__ == "__main__":
    main()



__version__ = "4.0.4"

from .helpers import determine_core_indices
from .rotations import proportional_orientations, quat_to_rotmat
from .structure import Structure
from .volume import Volume, structure_to_shape_like

__all__ = [
    "determine_core_indices",
    "proportional_orientations",
    "quat_to_rotmat",
    "Structure",
    "structure_to_shape_like",
    "Volume",
]



import csv
import gzip
import logging
import os
import shutil
from dataclasses import dataclass
from math import floor
from pathlib import Path
from string import Template
from textwrap import dedent
from typing import Any

from molviewspec import (
    MVSJ,
    GlobalMetadata,
    Snapshot,
    States,
    create_builder,
)
from molviewspec.builder import Representation, Root, VolumeRepresentation

from powerfit_em.volume import parse_volume

logger = logging.getLogger(__name__)


@dataclass
class Iso:
    value: float
    min: float
    max: float
    step: float


def _calc_rel_isovalue(volume_path: Path) -> Iso:
    """Calculate relative iso value from volume file.

    Args:
        volume_path: Path to volume file (.mrc or .map or .ccp4 optionally gzipped)

    Returns:
        A Iso object.
    """
    try:
        p = parse_volume(str(volume_path))
    except (KeyError, AttributeError):
        logger.warning("Could not determine relative iso value for density map, using default values.")
        return Iso(2, -10, 10, 0.1)
    # Logic taken from
    # https://github.com/molstar/molstar/blob/e4edb67f62cf92f31ce220f3d250bfd9c8f77572/src/mol-model/volume/volume.ts#L113-L127
    # and ccp4 header mapping from
    # https://github.com/molstar/molstar/blob/e4edb67f62cf92f31ce220f3d250bfd9c8f77572/src/mol-io/reader/ccp4/parser.ts#L70-L94
    amin = p.header["amin"]
    amax = p.header["amax"]
    amean = p.header["amean"]
    sigma = p.header["rms"]
    rel_min = (amin - amean) / sigma
    rel_max = (amax - amean) / sigma
    rel_isovalue = 2
    if sigma == 0:
        rel_isovalue = 0
    if rel_isovalue < rel_min:
        rel_isovalue = rel_min
    if rel_isovalue > rel_max:
        rel_isovalue = rel_max
    steps = 100
    step = round((amax - amin) / sigma / steps, 2)
    return Iso(rel_isovalue, rel_min, rel_max, step)


def _add_density_to_builder(builder: Root, density: Path, rel_iso_value: float) -> VolumeRepresentation:
    return (
        builder.download(url=density.name)
        .parse(format="map")
        .volume()
        .representation(type="isosurface", relative_isovalue=rel_iso_value, show_wireframe=True)
        .color(color="gray")
        .opacity(opacity=0.2)
    )


def _add_model_to_builder(builder: Root, model: Path, label: str | None = None) -> Representation:
    component = builder.download(url=model.name).parse(format="pdb").model_structure().component()
    if label:
        component = component.label(text=label)
    # Focus on fitted model with some of the density around it
    return component.focus(radius_factor=3).representation().color(color="blue")


def _create_snapshot_description(solution: dict[str, Any]) -> str:
    translation = f"({solution['x']}, {solution['y']}, {solution['z']})"
    rotation = f"(({solution['a11']}, {solution['a12']}, {solution['a13']}), "
    rotation += f"({solution['a21']}, {solution['a22']}, {solution['a23']}), "
    rotation += f"({solution['a31']}, {solution['a32']}, {solution['a33']}))"
    fitted_model_file = solution["fitted_model_file"]
    return dedent(f"""
            - Rank: {solution["rank"]}
            - Fitted model: [{fitted_model_file.name}]({fitted_model_file.name})
            - Cross correlation score: {solution["cc"]}
            - Fisher z-score: {solution["Fish-z"]}
            - Relative z-score (z-score/α): {solution["rel-z"]}
            - Sigma difference (z₁-zₙ/α): {solution["sigma_dif"]}
            - Translation: {translation}
            - Rotation: {rotation}
        """)


def create_snapshot(solution: dict, density: Path, rel_iso_value: float) -> Snapshot:
    builder = create_builder()
    fitted_model_file = solution["fitted_model_file"]
    _add_density_to_builder(builder, density, rel_iso_value)
    _add_model_to_builder(builder, fitted_model_file)
    description = _create_snapshot_description(solution)
    return builder.get_snapshot(
        key=fitted_model_file.stem,
        title=solution["rank"],
        description=description,
        description_format="markdown",
    )


def create_snapshot_with_all_models(
    fitted_model_files: list[Path],
    density: Path,
    rel_iso_value: float,
    max_models: int = 100,
) -> Snapshot:
    builder = create_builder()
    _add_density_to_builder(builder, density, rel_iso_value)
    for fitted_model_file in fitted_model_files[:max_models]:
        label = fitted_model_file.stem.replace("fit_", "")
        _add_model_to_builder(builder, fitted_model_file, label=label)
    maxed = len(fitted_model_files) > max_models
    description_first_line = f"First {max_models} fitted models." if maxed else "All fitted models."
    description = dedent(f"""
        {description_first_line}

        The labels correspond to the rank of the fitted model.

        Can take a while to load the models.
    """)
    return builder.get_snapshot(
        key="all_models",
        title=f"First {max_models} fitted models" if maxed else "All fitted models",
        description=description,
        description_format="markdown",
    )


def _read_solutions(path: Path, delimiter: str | None = None) -> list[dict]:
    if delimiter is None:
        raise ValueError("Delimiter must be set to produce a report.")
    with open(path) as f:
        reader = csv.DictReader(f, delimiter=delimiter)
        solutions = list(reader)
    # Calculate sigma_dif for each solution and add fitted_model_file
    if solutions:
        best_z = float(solutions[0]["rel-z"])
        for i, solution in enumerate(solutions):
            solution["sigma_dif"] = round(best_z - float(solution["rel-z"]), 3)
            solution["fitted_model_file"] = Path(path.parent) / f"fit_{i + 1}.pdb"
    return solutions


def generate_html(
    target_path: Path,
    iso: Iso,
    state_path: Path,
    options: dict[str, Any],
    solutions_table: str,
):
    # template was copied from molviewspec.molstar_widgets.STORIES_TEMPLATE and heavily modified
    template = Template(
        dedent("""\
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <title>PowerFit Report</title>
                <style>
                    * {
                        margin: 0;
                        padding: 0;
                        box-sizing: border-box;
                    }

                    body {
                        display: flex;
                        flex-direction: column;
                    }

                    header {
                        background: #F6F5F3;
                        padding: 1rem;
                        border-bottom: 1px solid #ccc;
                        display: flex;
                        justify-content: space-between;
                    }

                    #iso-number {
                        width: 4rem;
                    }

                    .options li {
                        list-style-position: inside;
                    }

                    .container {
                        flex: 1 1 auto;
                        display: flex;
                        flex-direction: row;
                        width: 100vw;
                        height: 600px;
                    }

                    #viewer {
                        flex: 2 1 0;
                        min-width: 0;
                        position: relative;
                        height: 100%;
                    }

                    #controls {
                        flex: 1 1 0;
                        min-width: 320px;
                        max-width: 600px;
                        padding: 16px;
                        padding-bottom: 20px;
                        border-left: 1px solid #ccc;
                        background: #F6F5F3;
                        display: flex;
                        flex-direction: column;
                        gap: 16px;
                        height: 100%;
                        box-sizing: border-box;
                    }

                    #solutions-table {
                        table {
                            border-collapse: collapse;
                            border: 1px solid #ccc;
                            margin: 1rem;

                            th, td {
                                border: 1px solid rgb(160 160 160);
                                padding: 8px 10px;
                            }
                        }
                        button {
                            display: inline-block;
                            padding: 0.25rem 0.5rem;
                            text-align: center;
                            font-size: 11px;
                            font-weight: 600;
                            letter-spacing: .1rem;
                            text-transform: uppercase;
                            text-decoration: none;
                            white-space: nowrap;
                            border-radius: 4px;
                            border: 1px solid #bbb;
                            cursor: pointer;
                            box-sizing: border-box
                        }
                    }

                    @media (orientation:portrait),
                    (max-width: 900px) {
                        .container {
                            flex-direction: column;
                        }

                        #viewer {
                            flex: 1 1 0;
                            min-height: 300px;
                            height: 50vh;
                        }

                        #controls {
                            flex: 1 1 0;
                            min-width: 0;
                            max-width: none;
                            border-left: none;
                            border-top: 1px solid #ccc;
                            height: auto;
                        }

                        .msp-viewport-controls-buttons {
                            display: none;
                        }
                    }
                </style>
                <script src="https://cdn.jsdelivr.net/npm/molstar@${molstar_version}/build/mvs-stories/mvs-stories.js"></script>
                <link rel="stylesheet" type="text/css"
                    href="https://cdn.jsdelivr.net/npm/molstar@${molstar_version}/build/mvs-stories/mvs-stories.css" />
            </head>

            <body>
                <header>
                    <h4>PowerFit Report</h4>
                    <details title="Options used for fitting">
                        <summary>Options</summary>
                        <span>The following options were used for fitting:</span>
                        <ul class="options">
                            ${options}
                        </ul>
                    </details>
                    <div>
                        Download:
                        <a href="${target}" target="_blank" rel="noreferrer">density map</a>
                        <a href="solutions.out" target="_blank" rel="noreferrer">solutions</a>
                        <a href="lcc.mrc" target="_blank" rel="noreferrer">cross-correlation</a>
                    </div>
                    <label title="Adjust the iso value of the density map.">
                        Iso value
                        <input id="iso-slider" type="range" min="${iso_min}" max="${iso_max}" value="${iso_value}"
                            step="${iso_step}" />
                        <input id="iso-number" type="number" min="${iso_min}" max="${iso_max}" value="${iso_value}"
                            step="${iso_step}" />
                    </label>
                </header>
                <div class="container">
                    <div id="viewer">
                        <mvs-stories-viewer />
                    </div>
                    <div id="controls">
                        <mvs-stories-snapshot-markdown style="flex-grow: 1;" />
                    </div>
                </div>
               <div id="solutions-table">
               ${solutions}
               </div>
                <script>
                    mvsStories.loadFromURL('${state}', { format: 'mvsj' });

                    let listenerAdded = false;
                    const slider = document.getElementById('iso-slider');
                    const number = document.getElementById('iso-number');
                    let currentIsoValue = parseFloat(number.value);

                    function updateIsoValue(newIsoValue) {
                        const context = mvsStories.getContext();
                        const viewer = context.state.viewers.value[0].model;
                        const plugin = viewer.plugin;
                        // Find and update all isosurface representations in the Mol* viewer
                        const state = viewer.plugin.state.data;
                        const isosurfaceNodes = Array.from(viewer.plugin.state.data.cells.values().filter(v => v.obj.label === 'Isosurface'))

                        // Update the relativeIsoValue for each isosurface
                        const b = state.build();

                        isosurfaceNodes.forEach(cell => {
                            const params = cell.params?.values;
                            if (!params?.type?.params?.isoValue) return;

                            const nextParams = {
                                ...params,
                                type: {
                                    ...params.type,
                                    params: {
                                        ...params.type.params,
                                        isoValue: {
                                            kind: 'relative',
                                            relativeValue: newIsoValue
                                        }
                                    }
                                }
                            };

                            b.to(cell.transform.ref).update(nextParams);
                        });
                        plugin.runTask(state.updateTree(b));

                        if (!listenerAdded) {
                            viewer.subscribe(viewer.plugin.managers.snapshot.events.changed, () => {
                                // wait a frame to apply iso so volume is present
                                requestAnimationFrame(() => {
                                    updateIsoValue(currentIsoValue);
                                });
                            });
                            listenerAdded = true;
                        }
                    }

                    slider.addEventListener('input', (event) => {
                        const newIsoValue = parseFloat(event.target.value);
                        if (newIsoValue === currentIsoValue) {
                            return;
                        }
                        currentIsoValue = newIsoValue;
                        updateIsoValue(currentIsoValue);
                        if (number.value !== event.target.value) {
                            number.value = currentIsoValue;
                        }
                    });
                    number.addEventListener('input', (event) => {
                        const newIsoValue = parseFloat(event.target.value);
                        if (newIsoValue === currentIsoValue) {
                            return;
                        }
                        currentIsoValue = newIsoValue;
                        updateIsoValue(currentIsoValue);
                        if (slider.value !== event.target.value) {
                            slider.value = currentIsoValue;
                        }
                    });

                    function jumpToSnapshot(key) {
                        const context = mvsStories.getContext();
                        const model = context.state.viewers.value[0].model;
                        model.plugin?.managers.snapshot.entryMap.forEach((entry, entryId) => {
                            if (entry.key === key) {
                                const snapshot = model.plugin?.managers.snapshot.setCurrent(entryId);
                                if (snapshot) {
                                    model.plugin.state.setSnapshot(snapshot);
                                }
                            }
                        })
                    }

                    function updateButtonStates(currentKey) {
                        // Enable all buttons first
                        document.querySelectorAll('#solutions-table button').forEach(button => {
                            button.disabled = false;
                            button.style.opacity = '1';
                            button.style.cursor = 'pointer';
                            button.title = 'View this solution in 3D viewer';
                        });

                        // Disable the button for the current snapshot
                        if (currentKey) {
                            const currentButton = document.querySelector(`#solutions-table button[data-snapshot-key="${currentKey}"]`);
                            if (currentButton) {
                                currentButton.disabled = true;
                                currentButton.style.opacity = '0.5';
                                currentButton.style.cursor = 'not-allowed';
                                currentButton.title = 'This solution is currently displayed';
                            }
                        }
                    }

                    function setupSnapshotListener() {
                        const context = mvsStories.getContext();
                        if (!context?.state?.viewers?.value?.[0]?.model) {
                            // Retry after a short delay if context is not ready
                            setTimeout(setupSnapshotListener, 100);
                            return;
                        }

                        const model = context.state.viewers.value[0].model;
                        const plugin = model.plugin;

                        if (plugin?.managers?.snapshot) {
                            // Update button states when snapshot changes
                            plugin.managers.snapshot.subscribe(plugin.managers.snapshot.events.changed, (newState) => {
                                const key = plugin.managers.snapshot.current.key;
                                updateButtonStates(key);
                            });

                            // Set initial state
                            const initialKey = plugin.managers.snapshot.current.key;
                            updateButtonStates(initialKey);
                        }
                    }
                    setupSnapshotListener();
                </script>
            </body>

            </html>
    """)  # noqa: E501
    )
    li_options = "\n".join([f"<li>{key}: {value}</li>" for key, value in options.items()])

    return template.safe_substitute(
        state=state_path.name,
        target=target_path.name,
        iso_min=round(iso.min, 2),
        iso_max=round(iso.max, 2),
        iso_value=round(iso.value, 2),
        iso_step=round(iso.step, 2),
        options=li_options,
        molstar_version="4.18.0",
        solutions=solutions_table,
    )


def generated_table(solutions: list[dict[str, Any]]) -> str:
    if not solutions:
        return "<p>No solutions found.</p>"
    table = dedent("""\
        <table>
        <thead>
          <tr>
            <th>Rank</th>
            <th>Cross correlation score</th>
            <th>Fisher z-score</th>
            <th>Relative z-score (z-score/&alpha;)</th>
            <th>Sigma difference (z<sub>1</sub>-z<sub>N</sub>/&alpha;)</th>
            <th></th>
          </tr>
        </thead>
        <tbody>
        """)
    green_gradient = (
        "#FFFFFF",
        "#F8FAF8",
        "#F2F5F2",
        "#EBF1EB",
        "#E5ECE5",
        "#DEE8DE",
        "#D8E3D8",
        "#D1DFD1",
        "#CBDACB",
        "#C4D5C4",
        "#BED1BE",
        "#B7CCB7",
        "#B1C8B1",
        "#AAC3AA",
        "#A4BFA4",
        "#9DBA9E",
        "#97B597",
        "#90B191",
        "#8AAC8A",
        "#83A884",
        "#7DA37D",
        "#769F77",
        "#709A70",
        "#69956A",
        "#639163",
        "#5C8C5D",
        "#568856",
        "#4F8350",
        "#497F49",
        "#427A43",
        "#3C763D",
    )
    # Improve contrast by using white text on dark green backgrounds
    invert_text_color_indices = {30}
    coloring_threshold = 3
    all_solutions_below_threshold = solutions[-1]["sigma_dif"] < coloring_threshold
    for solution in solutions:
        sigma_dif = solution["sigma_dif"]
        style = ""
        if sigma_dif < coloring_threshold and not all_solutions_below_threshold:
            gradient_index = min(30, max(0, floor(30 - 10 * sigma_dif)))
            color = green_gradient[gradient_index]
            if gradient_index in invert_text_color_indices:
                style = f"background-color: {color}; color: white; font-weight: 900;"
            else:
                style = f"background-color: {color}; font-weight: 900;"
        snapshot_key = f"fit_{solution['rank']}"
        fitted_model_file = solution["fitted_model_file"]
        table += dedent(f"""\
            <tr style="{style}">
              <td><a href="{fitted_model_file.name}" target="_blank" title="Download fitted model">{solution["rank"]}</a></td>
              <td>{solution["cc"]}</td>
              <td>{solution["Fish-z"]}</td>
              <td>{solution["rel-z"]}</td>
              <td>{sigma_dif}</td>
              <td><button data-snapshot-key="{snapshot_key}" onclick="jumpToSnapshot('{snapshot_key}')">View in 3D</button></td>
            </tr>
        """)  # noqa: E501
    table += "</tbody></table>\n"
    return table


def copy_target_to_report_dir(target: str, run_dir: Path) -> Path:
    target_path = run_dir / Path(target).name
    gunzip_needed = False
    if target_path.suffix == ".gz":
        target_path = target_path.with_suffix("")
        gunzip_needed = True
    if not target_path.exists():
        rtarget = Path(os.path.relpath(target, Path.cwd()))
        rrun_dir = Path(os.path.relpath(run_dir, Path.cwd()))
        if gunzip_needed:
            # Mol* stories can not render gzipped files.
            logger.warning(f"Uncompressing target file ({rtarget}) to report directory ({rrun_dir})")
            with gzip.open(target, "rb") as f_in, open(target_path, "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        else:
            logger.warning(f"Copying target file ({rtarget}) to report directory ({rrun_dir}).")
            shutil.copyfile(target, target_path)
    return target_path


def generate_report(
    directory: str,
    target: str,
    num: int,
    delimiter: str | None,
    options: dict[str, Any],
):
    """
    Generate a HTML report of the fitting results.

    The report uses Mol* and MolViewSpec stories.

    Args:
        directory: Directory containing the fitting results.
        target: Target volume file name.
        num: Number of fits to include in the report.
        delimiter: Delimiter used in the solutions file. If None, the report raise an error.
        options: Options used for fitting, to include in the report.
    """
    run_dir = Path(directory)

    # To render report all files need to be in the same directory
    target_path = copy_target_to_report_dir(target, run_dir)
    iso = _calc_rel_isovalue(target_path)

    solutions_file = run_dir / "solutions.out"
    solutions = _read_solutions(solutions_file, delimiter)
    fitted_model_files = [solution["fitted_model_file"] for solution in solutions[:num]]
    snapshots = []
    for solution in solutions[:num]:
        snapshot = create_snapshot(solution, target_path, iso.value)
        snapshots.append(snapshot)
    snapshots.append(create_snapshot_with_all_models(fitted_model_files, target_path, iso.value))
    # could add snapshot with lcc.mrc,
    # but could not make tutorial map look interesting,
    # so skip for now

    state = MVSJ(
        data=States(
            snapshots=snapshots,
            metadata=GlobalMetadata(
                title="",
                description="Fitted models and density map",
                description_format="markdown",
            ),
        )
    )
    state_path = run_dir / "state.mvsj"
    state_path.write_text(state.dumps(indent=2))

    solutions_table = generated_table(solutions[:num])

    report = run_dir / "report.html"
    body = generate_html(target_path, iso, state_path, options, solutions_table)
    report.write_text(body)

    rel_report = Path(os.path.relpath(report, Path.cwd()))
    rel_run_dir = Path(os.path.relpath(run_dir, Path.cwd()))
    logger.warning(
        f"Report generated at {rel_report}. Start web server with "
        f"`python3 -m http.server -d {rel_run_dir}`. "
        "Open http://localhost:8000/report.html in a web browser to view the results."
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=dedent("""\
            Generate report.html from fitting results.
                           
            Example usage:
            
            powerfit ribosome-KsgA.map 13 KsgA.pdb -d run -n 20 --delimiter , -a 20
            # Later generate report.html with
            python3 -m powerfit_em.report run ribosome-KsgA.map -n 20 --delimiter , --option resolution 13 --option angle 20
        """),  # noqa: E501
    )
    parser.add_argument(
        "directory",
        type=str,
        help="Directory with fitting results.",
    )
    parser.add_argument(
        "target",
        type=str,
        help="Target density map to fit the model in.",
    )
    parser.add_argument(
        "-n",
        "--num",
        dest="num",
        type=int,
        default=10,
        metavar="<int>",
        help="Number of models written to file. This number will be capped if less solutions are found as requested.",
    )
    parser.add_argument(
        "--delimiter",
        dest="delimiter",
        type=str,
        default=",",
        metavar="<str>",
        help="Delimiter used in the 'solutions.out' file. For example use ',' or '\\t'. Defaults to fixed width.",
    )
    parser.add_argument(
        "--option",
        action="append",
        nargs=2,
        metavar=("KEY", "VALUE"),
        default=[],
        help="Option used for fitting, to include in the report. "
        "Can be used multiple times. Example: --option resolution 13",
    )
    args = parser.parse_args()

    generate_report(args.directory, args.target, args.num, args.delimiter, dict(args.option))



from unittest import TestCase, main

import numpy as np

from powerfit_em._extensions import rotate_grid3d


class TestExtensions(TestCase):
    def test_rotate_grid3d(self):
        grid = np.zeros((4, 5, 6), dtype=np.float32)
        grid[0, 0, 0] = 1
        grid[0, 0, 1] = 1
        grid[0, 1, 1] = 1
        grid[0, 0, 2] = 1
        grid[0, 0, -1] = 1
        grid[-1, 0, 0] = 1
        # Identity rotation
        rotmat = np.asarray([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
        out = np.zeros_like(grid)
        rotate_grid3d(grid, rotmat, 2, out, True)
        self.assertTrue(np.allclose(out, grid))

        # 90 degree rotation around Z-axis
        out.fill(0)
        rotmat = np.asarray([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=np.float32)
        rotate_grid3d(grid, rotmat, 2, out, False)
        # Build answer
        answer = np.zeros_like(out)
        answer[0, 0, 0] = 1
        answer[0, 1, 0] = 1
        answer[0, 1, -1] = 1
        answer[0, 2, 0] = 1
        answer[0, -1, 0] = 1
        answer[-1, 0, 0] = 1
        self.assertTrue(np.allclose(answer, out))


if __name__ == "__main__":
    main()



import numpy as np
import numpy.linalg as la


class ShapeDescriptor:
    """Class that calculates rotation-invariant shape descriptors based on the
    radius of gyration tensor.
    """

    def __init__(self, coor):
        coor = np.asmatrix(coor)
        # Transpose coordinates if the shape is N, 3
        if coor.shape[1] == 3:
            coor = coor.T
        self.coor = coor - coor.mean(axis=1).reshape(-1, 1)
        self.gyration_tensor = self.coor * self.coor.T
        eig_values, eig_vectors = la.eigh(self.gyration_tensor)
        # Sort eigenvalues so that l_x <= l_y <= l_z.
        sort_ind = np.argsort(eig_values)
        self.principal_moments = eig_values[sort_ind]
        self.lx, self.ly, self.lz = self.principal_moments
        self.principal_axes = eig_vectors[:, sort_ind]

    @property
    def Rg2(self):
        """Return the radius of gyration squared (Rg^2)"""
        return self.principal_moments.sum()

    @property
    def asphericity(self):
        return (3 * self.lz - self.Rg2) / 2

    @property
    def acylindricity(self):
        return self.ly - self.lx

    @property
    def relative_shape_anisotropy(self):
        """Returns the relatve shape anisotropy in the interval [0, 1].

        0 for spherical conformation
        1 for for linear chains.

        See Arkin and Janke. Journal of Chemical Physics 138, 2013.
        """
        lx, ly, lz = self.principal_moments
        return 1 - 3 * (lx * ly + ly * lz + lz * lx) / self.Rg2**2

    @property
    def shape_anisotropy(self):
        ave_moment = self.principal_moments.mean()
        return np.prod(self.principal_moments - ave_moment) / ave_moment**3



import warnings
from functools import partial

import numpy as np
from numpy import typing as npt
from scipy.ndimage import laplace as laplace_filter
from tqdm import tqdm

from powerfit_em._extensions import rotate_grid3d
from powerfit_em.correlators.shared import Correlator, Vars, VarsFT, f32, get_ft_shape, get_lcc_mask, i32
from powerfit_em.helpers import pyfftw_available


def build_ffts(target: np.ndarray, gcc: np.ndarray, ft_gcc: np.ndarray, fftw: bool):
    """Build the FFTs (in case of pyfftw), or patch the numpy fft interface to resemble pyfftw."""
    if fftw and pyfftw_available():
        from pyfftw.builders import irfftn as irfftn_builder
        from pyfftw.builders import rfftn as rfftn_builder

        rfftn = rfftn_builder(gcc)
        irfftn = irfftn_builder(ft_gcc, s=target.shape)
    else:
        warnings.warn("Using numpy for calculating score. Install pyFFTW for faster calculation.")  # noqa: B028

        def rfftn(src: np.ndarray, dst: np.ndarray):
            np.fft.rfftn(src, out=dst)

        def irfftn(src: np.ndarray, dst: np.ndarray):
            np.fft.irfft(src, out=dst)

    return rfftn, irfftn


def rmax(target: np.ndarray) -> int:
    return min(target.shape) // 2


def zeros_array(shape: tuple[int], dtype: npt.DTypeLike, fftw: bool) -> np.ndarray:
    """Returns optimally SIMD aligned array if PyFFTW is used, for faster computation."""
    if fftw and pyfftw_available():
        from pyfftw import simd_alignment, zeros_aligned

        return zeros_aligned(shape, dtype, n=simd_alignment)
    else:
        return np.zeros(shape, dtype)


def init_cpu_vars(
    target: np.ndarray,
    laplace: bool,
    fftw: bool,
) -> tuple[Vars[np.ndarray, np.ndarray], VarsFT[np.ndarray]]:
    """Initialize all CPU variables on the specified queue."""

    lcc_mask = get_lcc_mask(target)
    _t = laplace_filter(target, mode="wrap") if laplace else target

    vars = Vars(
        target=_t.astype(f32),
        template=zeros_array(target.shape, f32, fftw),
        mask=zeros_array(target.shape, f32, fftw),
        lcc_mask=lcc_mask.astype(np.uint8),
        target2=zeros_array(target.shape, f32, fftw),
        rot_template=zeros_array(target.shape, f32, fftw),
        rot_mask=zeros_array(target.shape, f32, fftw),
        rot_mask2=zeros_array(target.shape, f32, fftw),
        gcc=zeros_array(target.shape, f32, fftw),
        ave=zeros_array(target.shape, f32, fftw),
        ave2=zeros_array(target.shape, f32, fftw),
        lcc=zeros_array(target.shape, f32, fftw),
        rot=zeros_array(target.shape, i32, fftw),
    )

    vars_ft = VarsFT(
        target=zeros_array(get_ft_shape(target), np.complex64, fftw),
        target2=zeros_array(get_ft_shape(target), np.complex64, fftw),
        template=zeros_array(get_ft_shape(target), np.complex64, fftw),
        mask=zeros_array(get_ft_shape(target), np.complex64, fftw),
        mask2=zeros_array(get_ft_shape(target), np.complex64, fftw),
        ave=zeros_array(get_ft_shape(target), np.complex64, fftw),
        ave2=zeros_array(get_ft_shape(target), np.complex64, fftw),
        lcc=zeros_array(get_ft_shape(target), np.complex64, fftw),
        gcc=zeros_array(get_ft_shape(target), np.complex64, fftw),
    )
    return vars, vars_ft


class CPUCorrelator(Correlator):
    """Compute the LCC score for a target and template combination."""

    def __init__(
        self,
        target: np.ndarray,
        template: np.ndarray,
        rotations: np.ndarray,
        mask: np.ndarray,
        laplace: bool = False,
        fftw: bool = True,
    ):
        """Initialize the CPU correlator.

        Args:
            target: the target density on which you want to fit a template structure.
            template: the template structure that you want to fit in the target density,
                should have been regridded to the same grid as the target density.
            rotations: array of 3D-rotation matrices, of shape (n_rotations, 3, 3).
                for each rotation the local cross correlation is computed for every
                possible translation (in parallel using Fourier transforms).
            mask: core-weighted mask. See doi:10.3934/biophy.2015.2.73, Figure 1.
            laplace: if true, a Laplace pre-filter is applied to the target density and
                template to enhance the sensitivity of the scoring function.
            fftw: if true, the PyFFTW library will be used to compute the fourier transforms.
        """
        self.target: np.ndarray = target / target.max()
        self.laplace = laplace
        self.rotations = rotations

        self.vars, self.vars_ft = init_cpu_vars(self.target, self.laplace, fftw)

        self.lcc_scan = np.zeros(self.target.shape, dtype=f32)
        self.lcc = np.zeros(self.target.shape, dtype=f32)
        self.rot = np.zeros(self.target.shape, dtype=i32)

        self.set_template(template, mask)

        # set methods in the same was as GPUCorrelator implementation;
        self.conj_multiply = lambda a, b, c: np.multiply(np.conjugate(a), b, out=c)
        self.square = lambda a, b: np.square(a, out=b)
        self.rfftn, self.irfftn = build_ffts(self.target, self.vars.gcc, self.vars_ft.gcc, fftw)

        # pre-calculate the FFTs of the target
        self.rfftn(self.vars.target, self.vars_ft.target)
        self.rfftn(self.vars.target**2, self.vars_ft.target2)

    def _set_template_var(self, template: np.ndarray):
        self.vars.template = template.astype(f32)

    def _set_mask_var(self, mask: np.ndarray):
        self.vars.mask = mask.astype(f32)

    def rotate_grids(self, rotmat: np.ndarray):
        """Rotate the template and mask using the rotational matrix."""
        rotate_grid3d(self.vars.template, rotmat.astype(f32), rmax(self.target), self.vars.rot_template, False)
        rotate_grid3d(self.vars.mask, rotmat.astype(f32), rmax(self.target), self.vars.rot_mask, True)

    def compute_lcc_score_and_take_best(self, n: int):
        """Compute the LCC score and store best result.

        Args:
            n: iteration number.
        """
        self.vars.ave2 *= self.norm_factor

        var = self.vars.ave2 - self.vars.ave**2
        # Ignore division and invalid sqrt errors
        #    see: https://github.com/haddocking/powerfit/issues/72
        with np.errstate(divide="ignore", invalid="ignore"):
            self.lcc_scan = np.where(self.vars.lcc_mask, self.vars.gcc / np.sqrt(var), 0.0)
        ind = np.greater(self.lcc_scan, self.lcc)
        # store lcc and rotation index
        self.lcc[ind] = self.lcc_scan[ind]
        self.rot[ind] = n

    def scan(self, progress: partial[tqdm] | None = None):
        """Scan all provided rotations to find the best fit."""
        self.vars.lcc.fill(0)
        self.vars.rot.fill(0)

        _range = range(0, self.rotations.shape[0])
        if progress is not None:
            _range = progress(_range)

        for n in _range:
            self.compute_rotation(n, self.rotations[n])



import importlib.resources
from math import sqrt
from random import random

import numpy as np

import powerfit_em.data


def euler(angle, axis):
    cos_a = np.cos(angle)
    sin_a = np.sin(angle)
    if axis == "x":
        matrix = [[1, 0, 0], [0, cos_a, -sin_a], [0, sin_a, cos_a]]
    elif axis == "y":
        matrix = [[cos_a, 0, sin_a], [0, 1, 0], [-sin_a, 0, cos_a]]
    elif axis == "z":
        matrix = [[cos_a, -sin_a, 0], [sin_a, cos_a, 0], [0, 0, 1]]
    else:
        raise ValueError("Axis is not recognized.")
    return np.asmatrix(matrix, dtype=np.float64)


def euler_to_rotmat(angles, order="zyz"):
    rotmat = np.asmatrix(np.identity(3, dtype=np.float64))
    for angle, axis in zip(angles, order, strict=False):
        rotmat *= euler(angle, axis)
    return rotmat


def quat_to_rotmat(quaternions):
    w = quaternions[:, 0]
    x = quaternions[:, 1]
    y = quaternions[:, 2]
    z = quaternions[:, 3]

    Nq = w**2 + x**2 + y**2 + z**2
    s = np.zeros(Nq.shape, dtype=np.float64)
    s[Nq > 0.0] = 2.0 / Nq[Nq > 0.0]
    s[Nq <= 0.0] = 0

    X = x * s
    Y = y * s
    Z = z * s

    rotmat = np.zeros((quaternions.shape[0], 3, 3), dtype=np.float64)
    rotmat[:, 0, 0] = 1.0 - (y * Y + z * Z)
    rotmat[:, 0, 1] = x * Y - w * Z
    rotmat[:, 0, 2] = x * Z + w * Y

    rotmat[:, 1, 0] = x * Y + w * Z
    rotmat[:, 1, 1] = 1.0 - (x * X + z * Z)
    rotmat[:, 1, 2] = y * Z - w * X

    rotmat[:, 2, 0] = x * Z - w * Y
    rotmat[:, 2, 1] = y * Z + w * X
    rotmat[:, 2, 2] = 1.0 - (x * X + y * Y)

    np.around(rotmat, decimals=8, out=rotmat)

    return rotmat


def random_rotmat():
    """Return a random rotation matrix"""

    s1 = 1
    while s1 >= 1.0:
        e1 = random() * 2 - 1
        e2 = random() * 2 - 1
        s1 = e1**2 + e2**2

    s2 = 1
    while s2 >= 1.0:
        e3 = random() * 2 - 1
        e4 = random() * 2 - 1
        s2 = e3**2 + e4**2

    q0 = e1
    q1 = e2
    q2 = e3 * sqrt((1 - s1) / s2)
    q3 = e4 * sqrt((1 - s1) / s2)

    quat = [[q0, q1, q2, q3]]
    return quat_to_rotmat(np.asarray(quat))[0]


def proportional_orientations(angle):
    # orientation sets available: name of file: (Norientations, degree)
    rot_sets = {
        "E.npy": (1, 360.0),
        "c48u1.npy": (24, 62.8),
        "c600v.npy": (60, 44.48),
        "c48n9.npy": (216, 36.47),
        "c600vc.npy": (360, 27.78),
        "c48u27.npy": (648, 20.83),
        "c48u83.npy": (1992, 16.29),
        "c48u181.npy": (4344, 12.29),
        "c48n309.npy": (7416, 9.72),
        "c48n527.npy": (12648, 8.17),
        "c48u815.npy": (19560, 7.4),
        "c48u1153.npy": (27672, 6.6),
        "c48u1201.npy": (28824, 6.48),
        "c48u1641.npy": (39384, 5.75),
        "c48u2219.npy": (53256, 5.27),
        "c48u2947.npy": (70728, 4.71),
        "c48u3733.npy": (89592, 4.37),
        "c48u4749.npy": (113976, 4.0),
        "c48u5879.npy": (141096, 3.74),
        "c48u7111.npy": (170664, 3.53),
        "c48u8649.npy": (207576, 3.26),
    }

    # determine the apropiate set to use
    smallestdiff = None
    for s, n in rot_sets.items():
        alpha = n[1]
        diff = abs(angle - alpha)
        if smallestdiff is None or diff < smallestdiff:
            smallestdiff = diff
            fname = s

    # read file
    infile = importlib.resources.open_binary(powerfit_em.data, fname)
    quat_weights = np.load(infile)

    quat = quat_weights[:, :4]
    weights = quat_weights[:, -1]
    alpha = rot_sets[fname][1]

    return quat, weights, alpha



#! env/bin/python
import os.path

import numpy
from Cython.Build import cythonize
from setuptools import setup
from setuptools.extension import Extension


def main():
    ext_modules = [
        Extension(
            "powerfit_em._powerfit",
            [os.path.join("src", "powerfit_em", "_powerfit.pyx")],
            include_dirs=[numpy.get_include()],
        ),
        Extension(
            "powerfit_em._extensions",
            [os.path.join("src", "powerfit_em", "_extensions.c")],
            include_dirs=[numpy.get_include()],
            extra_compile_args=["-ffast-math"],
        ),
    ]

    ext_modules = cythonize(ext_modules)

    setup(
        ext_modules=ext_modules,
    )


if __name__ == "__main__":
    main()



import gzip
from pathlib import Path

import numpy as np
import numpy.testing as npt
import pytest

from powerfit_em.volume import Volume


@pytest.fixture
def example_volume() -> Volume:
    array = np.zeros((3, 4, 5))
    array[0, 0, 0] = 1.1
    array[1, 2, 3] = 2.2
    array[2, 3, 4] = 3.3
    return Volume(array)


@pytest.fixture
def example_mrc_file(tmp_path: Path, example_volume: Volume) -> Path:
    fn = tmp_path / "example_volume.mrc"
    example_volume.tofile(fn)
    return fn


@pytest.fixture
def example_gz_file(tmp_path: Path, example_mrc_file: Path) -> Path:
    with open(example_mrc_file, mode="rb") as f:
        mrc_file = f.read()

    fn = tmp_path / "example_volume.mrc.gz"
    with gzip.open(fn, "wb") as f:
        f.write(mrc_file)
    return fn


def test_to_file(example_mrc_file: Path):
    assert example_mrc_file.exists()
    assert example_mrc_file.stat().st_size > 0


def test_shape(example_volume: Volume):
    assert example_volume.shape == (3, 4, 5)


def test_dimensions(example_volume: Volume):
    npt.assert_array_equal(example_volume.dimensions, (5.0, 4.0, 3.0))


def test_get_start(example_volume: Volume):
    npt.assert_array_equal(example_volume.start, (0, 0, 0))


def test_fromfile(example_mrc_file: Path):
    volume = Volume.fromfile(str(example_mrc_file))

    assert volume.shape == (3, 4, 5)
    npt.assert_array_equal(volume.start, (0, 0, 0))
    npt.assert_array_equal(volume.dimensions, (5.0, 4.0, 3.0))
    expected = np.zeros((3, 4, 5))
    expected[0, 0, 0] = 1.1
    expected[1, 2, 3] = 2.2
    expected[2, 3, 4] = 3.3
    npt.assert_array_almost_equal(volume.array, expected)


def test_read_gzip(example_mrc_file: Path, example_gz_file: Path):
    mrc = Volume.fromfile(str(example_mrc_file))
    mrcgz = Volume.fromfile(str(example_gz_file))
    assert mrc == mrcgz



import gzip
import os.path
import warnings
from io import BufferedReader
from pathlib import Path
from struct import pack as _pack
from struct import unpack as _unpack
from sys import byteorder as _BYTEORDER

import numpy as np
from scipy.ndimage import gaussian_filter, zoom

from ._powerfit import blur_points, dilate_points


class Volume:
    @classmethod
    def fromfile(cls, fid, fmt=None):
        p = parse_volume(fid, fmt)
        return cls(p.density, p.voxelspacing, p.origin)

    def __init__(self, array, voxelspacing=1.0, origin=(0, 0, 0)):
        self.array = array
        self.voxelspacing = voxelspacing
        self.origin = origin

    def __eq__(self, other):
        if isinstance(other, Volume):
            return (
                (self.array == other.array).all()
                and self.voxelspacing == other.voxelspacing
                and self.origin == other.origin
            )
        else:
            return NotImplemented

    @property
    def shape(self):
        return self.array.shape

    @property
    def dimensions(self):
        return np.asarray([x * self.voxelspacing for x in self.array.shape][::-1])

    @property
    def start(self):
        return np.asarray([x / self.voxelspacing for x in self.origin])

    @start.setter
    def start(self, start):
        self._origin = np.asarray([x * self.voxelspacing for x in start])

    def duplicate(self):
        return Volume(self.array.copy(), voxelspacing=self.voxelspacing, origin=self.origin)

    def tofile(self, fid, fmt=None):
        if fmt is None:
            fmt = os.path.splitext(fid)[-1][1:]
        if fmt in ("ccp4", "map", "mrc"):
            to_mrc(fid, self)
        elif fmt in ("xplor", "cns"):
            to_xplor(fid, self)
        else:
            raise RuntimeError("Format is not supported.")


# builders
def zeros(shape, voxelspacing, origin):
    return Volume(np.zeros(shape), voxelspacing, origin)


def zeros_like(volume):
    return Volume(np.zeros_like(volume.array), volume.voxelspacing, volume.origin)


def resample(volume, factor, order=1):
    # suppress zoom UserWarning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        resampled_array = zoom(volume.array, factor, order=order)
    new_voxelspacing = volume.voxelspacing / factor
    return Volume(resampled_array, new_voxelspacing, volume.origin)


def trim(volume, cutoff, margin=2):
    if volume.array.max() <= cutoff:
        raise ValueError("Cutoff value should be lower than density max.")

    extent = []
    for axis in range(volume.array.ndim):
        tmp = np.swapaxes(volume.array, 0, axis)
        for n, s in enumerate(tmp):
            if s.max() > cutoff:
                low = max(0, n - margin)
                break
        for n, s in enumerate(tmp[::-1]):
            if s.max() > cutoff:
                high = min(tmp.shape[0], tmp.shape[0] - n + margin)
                break
        extent.append(slice(low, high))
    sub_array = volume.array[tuple(extent)]
    origin = [
        coor_origin + volume.voxelspacing * ext.start
        for coor_origin, ext in zip(volume.origin, extent[::-1], strict=False)
    ]
    return Volume(sub_array, volume.voxelspacing, origin)


def extend(volume, shape):
    new_volume = zeros(shape, volume.voxelspacing, volume.origin)
    ind = [slice(x) for x in volume.shape]
    new_volume.array[tuple(ind)] = volume.array
    return new_volume


def nearest_multiple2357(num):
    nearest = num
    while not is_multiple2357(nearest):
        nearest += 1
    return nearest


def is_multiple2357(num):
    """Returns the nearest larger number that is a multiple of 2, 3, and 5"""

    MULTIPLES = (2, 3, 5, 7)
    for multiple in MULTIPLES:
        while divmod(num, multiple)[1] == 0:
            num /= multiple
    return num == 1


def res_to_sigma(resolution):
    return resolution / (np.sqrt(2.0) * np.pi)


def sigma_to_res(sigma):
    return sigma * (np.sqrt(2.0) * np.pi)


def lower_resolution(vol, res_high, res_low):
    """Lowers the resolution of the volume"""

    # the sigma of the Gaussian kernel that will lower the density with the
    # specified amount is given Sk = sqrt(Sn^2 - Sc^2)
    # where Sk is the sigma of the kernel, Sn is the new sigma,
    # and Sc is the current sigma
    # See http://mathworld.wolfram.com/Convolution.html
    sigma_high = res_to_sigma(res_high)
    sigma_low = res_to_sigma(res_low)
    sigma_k = np.sqrt(sigma_low**2 - sigma_high**2) / vol.voxelspacing
    blurred_array = gaussian_filter(vol.array, sigma_k, mode="constant")
    return Volume(blurred_array, vol.voxelspacing, vol.origin)


def structure_to_shape(xyz, resolution, out=None, voxelspacing=None, radii=None, weights=None, shape="vol"):
    if shape not in ("vol", "mask"):
        raise ValueError("shape should either be 'vol' or 'mask'")

    if out is None and voxelspacing is None:  # noqa: SIM108
        voxelspacing = resolution / 4.0
    else:
        voxelspacing = out.voxelspacing

    if shape == "vol":
        if weights is None:
            weights = np.ones(xyz.shape[1])
        elif weights.size != xyz.shape[1]:
            raise ValueError("weights array is of incorrect size")

    if shape == "mask":
        if radii is None:
            radii = np.empty(xyz.shape[1], dtype=np.float64)
            radii.fill(5)
        elif radii.size != xyz.shape[1]:
            raise ValueError("weights array is of incorrect size")
        radii /= voxelspacing

    sigma = res_to_sigma(resolution)
    if out is None:
        extend = 4 * sigma
        imin = xyz.min(axis=1) - extend
        imax = xyz.max(axis=1) + extend
        center = (imin + imax) / 2.0
        ni = (np.ceil((imax - imin) / voxelspacing)).astype(np.int32)
        origin = center - (ni * voxelspacing) / 2.0
        grid = np.zeros(ni[::-1])
        xyz_grid = xyz + (ni * voxelspacing / 2.0 - center).reshape(-1, 1)
        out = Volume(grid, voxelspacing, origin)
    else:
        xyz_grid = xyz - out.origin.reshape(-1, 1)
    xyz_grid /= voxelspacing

    if shape == "vol":
        blur_points(xyz_grid, weights, sigma / voxelspacing, out.array, True)
    elif shape == "mask":
        dilate_points(xyz_grid, radii, out.array, True)
    return out


def structure_to_shape_like(vol, xyz, resolution=None, weights=None, radii=None, shape="vol"):
    if resolution is None:
        resolution = vol.resolution

    if shape == " vol":
        if weights is None:
            weights = np.ones(xyz.shape[1])
        elif weights.size != xyz.shape[1]:
            raise ValueError("weights array is of incorrect size")

    if shape == "mask":
        if radii is None:
            radii = np.empty(xyz.shape[1], dtype=np.float64)
            # radii.fill(0.5 * resolution)
            radii.fill(5)
        elif radii.size != xyz.shape[1]:
            raise ValueError("weights array is of incorrect size")
        radii /= vol.voxelspacing

    sigma = (resolution / (np.sqrt(2.0) * np.pi)) / vol.voxelspacing

    # move the coordinates to the origin of the grid
    xyz_grid = xyz - np.asarray(vol.origin, dtype=np.float64).reshape(3, 1)
    xyz_grid /= vol.voxelspacing

    out = zeros_like(vol)
    if shape == "vol":
        blur_points(xyz_grid, weights, sigma, out.array, True)
    elif shape == "mask":
        dilate_points(xyz_grid, radii, out.array, True)
    return out


# Volume parsers
def parse_volume(fid, fmt=None) -> "CCP4Parser | MRCParser | XPLORParser":
    try:
        fname = fid.name
    except AttributeError:
        fname = fid

    if fmt is None:
        fp = Path(fname)
        fmt = fp.suffix
        if fmt == ".gz":
            fmt = fp.suffixes[-2] + fmt

    if fmt in (".ccp4", ".map"):
        return CCP4Parser(fid)
    elif fmt in (".ccp4.gz", ".map.gz"):
        return CCP4Parser(fid, gzipped=True)
    elif fmt == ".mrc":
        return MRCParser(fname)
    elif fmt == ".mrc.gz":
        return MRCParser(fname, gzipped=True)
    elif fmt in (".xplor", ".cns"):
        return XPLORParser(fname)
    else:
        raise ValueError("Extension of file is not supported.")


class CCP4Parser:
    HEADER_SIZE = 1024
    HEADER_TYPE = "i" * 10 + "f" * 6 + "i" * 3 + "f" * 3 + "i" * 3 + "f" * 27 + "c" * 8 + "f" * 1 + "i" * 1 + "c" * 800
    HEADER_FIELDS = [
        "nc",
        "nr",
        "ns",
        "mode",
        "ncstart",
        "nrstart",
        "nsstart",
        "nx",
        "ny",
        "nz",
        "xlength",
        "ylength",
        "zlength",
        "alpha",
        "beta",
        "gamma",
        "mapc",
        "mapr",
        "maps",
        "amin",
        "amax",
        "amean",
        "ispg",
        "nsymbt",
        "lskflg",
        "skwmat",
        "skwtrn",
        "extra",
        "xstart",
        "ystart",
        "zstart",
        "map",
        "machst",
        "rms",
        "nlabel",
        "label",
    ]
    HEADER_CHUNKS = [1] * 25 + [9, 3, 12] + [1] * 3 + [4, 4, 1, 1, 800]

    def __init__(self, fid, gzipped: bool = False):
        if isinstance(fid, str):
            if gzipped:  # noqa: SIM108
                fhandle = gzip.open(fid, mode="rb")  # noqa: SIM115
            else:
                fhandle = open(fid, mode="rb")  # noqa: SIM115
        elif isinstance(fid, BufferedReader):
            fhandle = gzip.GzipFile(fileobj=fid, mode="rb") if gzipped else fid
        else:
            raise ValueError("Input should either be a file or filename.")

        self.fhandle = fhandle
        self.fname = fhandle.name

        # first determine the endiannes of the file
        self._get_endiannes()
        # get the header
        self._get_header()
        # Symmetry and non-rectangular boxes are not supported.
        is_orthogonal = True
        for angle_name in ["alpha", "beta", "gamma"]:
            angle = self.header[angle_name]
            if abs(angle - 90) > 1e-3:
                is_orthogonal = False
                break
        if not is_orthogonal:
            msg = "Only densities in rectangular boxes are supported."
            raise RuntimeError(msg)

        # check the order of axis in the file
        self._get_order()
        # determine the voxelspacing and origin
        spacings = []
        for axis_name in "xyz":
            length = self.header[axis_name + "length"]
            nvoxels = self.header["n" + axis_name]
            spacing = length / float(nvoxels)
            spacings.append(spacing)

        equal_spacing = True
        average = sum(spacings) / float(len(spacings))
        for spacing in spacings:
            if abs(spacing - average) > 1e-4:
                equal_spacing = False
        if not equal_spacing:
            msg = "Voxel spacing is not equal in all directions."
            raise RuntimeError(msg)

        self.voxelspacing = spacings[0]
        self.origin = self._get_origin()
        # generate the density
        shape_fields = ["nz", "ny", "nx"]
        self.shape = [self.header[field] for field in shape_fields]
        self._get_density(gzipped)

    def _get_endiannes(self):
        self.fhandle.seek(212)
        m_stamp = hex(ord(self.fhandle.read(1)))
        if m_stamp == "0x44":
            endian = "<"
        elif m_stamp == "0x11":
            endian = ">"
        else:
            raise RuntimeError("Endiannes is not properly set in file. Check the file format.")
        self._endian = endian
        self.fhandle.seek(0)

    def _get_header(self):
        header = _unpack(self._endian + self.HEADER_TYPE, self.fhandle.read(self.HEADER_SIZE))
        self.header = {}
        index = 0
        for field, nchunks in zip(self.HEADER_FIELDS, self.HEADER_CHUNKS, strict=False):
            end = index + nchunks
            if nchunks > 1:
                self.header[field] = header[index:end]
            else:
                self.header[field] = header[index]
            index = end
        self.header["label"] = "".join([c.decode("ascii") for c in self.header["label"]])
        self.header["map"] = "".join([c.decode("ascii") for c in self.header["map"]])
        self.header["machst"] = "".join([c.decode("ascii") for c in self.header["machst"]])

    def _get_origin(self):
        start_fields = ["nsstart", "nrstart", "ncstart"]
        start = [self.header[field] for field in start_fields]
        # Take care of axis order
        start = [start[x - 1] for x in self.order]
        return np.asarray([x * self.voxelspacing for x in start])

    def _get_density(self, gzipped: bool):
        # Determine the dtype of the file based on the mode
        mode = self.header["mode"]
        if mode == 0:
            dtype = "i1"
        elif mode == 1:
            dtype = "i2"
        elif mode == 2:
            dtype = "f4"
        else:
            msg = f"Unknown mode '{mode}' in file header."
            raise ValueError(msg)

        if gzipped:
            density = np.frombuffer(self.fhandle.read(), dtype=self._endian + dtype).reshape(self.shape)
        else:
            density = np.fromfile(self.fhandle, dtype=self._endian + dtype).reshape(self.shape)
        if self.order == (1, 3, 2):
            self.density = np.swapaxes(self.density, 0, 1)
        elif self.order == (2, 1, 3):
            self.density = np.swapaxes(self.density, 1, 2)
        elif self.order == (2, 3, 1):
            self.density = np.swapaxes(self.density, 2, 0)
            self.density = np.swapaxes(self.density, 0, 1)
        elif self.order == (3, 1, 2):
            self.density = np.swapaxes(self.density, 2, 1)
            self.density = np.swapaxes(self.density, 0, 2)
        elif self.order == (3, 2, 1):
            self.density = np.swapaxes(self.density, 0, 2)

        # Upgrade precision to double if float, and to int32 if int16
        if mode == 1:
            density = density.astype(np.int32)
        elif mode == 2:
            density = density.astype(np.float64)
        self.density = density

    def _get_order(self):
        self.order = tuple(self.header[axis] for axis in ("mapc", "mapr", "maps"))


class MRCParser(CCP4Parser):
    def _get_origin(self):
        origin_fields = ["xstart", "ystart", "zstart"]
        origin = [self.header[field] for field in origin_fields]
        return origin


def to_mrc(fid, volume, labels=[], fmt=None):  # noqa: B006
    if fmt is None:
        fmt = os.path.splitext(fid)[-1][1:]

    if fmt not in ("ccp4", "mrc", "map"):
        raise ValueError("Format is not recognized. Use ccp4, mrc, or map.")

    nz, ny, nx = volume.shape
    dtype = volume.array.dtype.name
    if dtype == "int8":
        mode = 0
    elif dtype in ("int16", "int32"):
        mode = 1
    elif dtype in ("float32", "float64"):
        mode = 2
    else:
        raise TypeError(f"Data type ({dtype})is not supported.")
    if fmt in ("ccp4", "map"):
        nxstart, nystart, nzstart = [int(round(x)) for x in volume.start]
    else:
        nxstart, nystart, nzstart = [0, 0, 0]
    xl, yl, zl = volume.dimensions
    alpha = beta = gamma = 90.0
    mapc, mapr, maps = [1, 2, 3]
    ispg = 1
    nsymbt = 0
    lskflg = 0
    skwmat = [0.0] * 9
    skwtrn = [0.0] * 3
    fut_use = [0.0] * 12
    if fmt == "mrc":  # noqa: SIM108
        origin = volume.origin
    else:
        origin = [0, 0, 0]
    str_map = b"MAP "
    if _BYTEORDER == "little" or _BYTEORDER == "big":
        machst = b"\x44\x41\x00\x00"
    else:
        raise ValueError(f"Byteorder {_BYTEORDER} is not recognized")
    labels = b"".join([b" "] * 800)
    nlabels = 0
    min_density = volume.array.min()
    max_density = volume.array.max()
    mean_density = volume.array.mean()
    std_density = volume.array.std()

    with open(fid, "wb") as out:
        out.write(_pack("i", nx))
        out.write(_pack("i", ny))
        out.write(_pack("i", nz))
        out.write(_pack("i", mode))
        out.write(_pack("i", nxstart))
        out.write(_pack("i", nystart))
        out.write(_pack("i", nzstart))
        out.write(_pack("i", nx))
        out.write(_pack("i", ny))
        out.write(_pack("i", nz))
        out.write(_pack("f", xl))
        out.write(_pack("f", yl))
        out.write(_pack("f", zl))
        out.write(_pack("f", alpha))
        out.write(_pack("f", beta))
        out.write(_pack("f", gamma))
        out.write(_pack("i", mapc))
        out.write(_pack("i", mapr))
        out.write(_pack("i", maps))
        out.write(_pack("f", min_density))
        out.write(_pack("f", max_density))
        out.write(_pack("f", mean_density))
        out.write(_pack("i", ispg))
        out.write(_pack("i", nsymbt))
        out.write(_pack("i", lskflg))
        for f in skwmat:
            out.write(_pack("f", f))
        for f in skwtrn:
            out.write(_pack("f", f))
        for f in fut_use:
            out.write(_pack("f", f))
        for f in origin:
            out.write(_pack("f", f))
        out.write(str_map)
        out.write(machst)
        out.write(_pack("f", std_density))
        # max 10 labels
        # nlabels = min(len(labels), 10)
        # TODO labels not handled correctly
        # for label in labels:
        #     list_label = [c for c in label]
        #     llabel = len(list_label)
        #     if llabel < 80:
        #
        #     # max 80 characters
        #     label = min(len(label), 80)
        out.write(_pack("i", nlabels))
        out.write(labels)
        # write density
        modes = [np.int8, np.int16, np.float32]
        volume.array.astype(modes[mode]).tofile(out)


class XPLORParser:
    """
    Class for reading XPLOR volume files created by NIH-XPLOR or CNS.
    """

    def __init__(self, fid):
        if isinstance(fid, BufferedReader):
            fname = fid.name
        elif isinstance(fid, str):
            fname = fid
            fid = open(fid)  # noqa: SIM115
        else:
            raise TypeError("Input should either be a file or filename")

        self.source = fname
        self._get_header()

    def _get_header(self):
        header = {}
        with open(self.source) as volume:
            # first line is blank
            volume.readline()

            line = volume.readline()
            nlabels = int(line.split()[0])

            label = [volume.readline() for n in range(nlabels)]
            header["label"] = label

            line = volume.readline()
            header["nx"] = int(line[0:8])
            header["nxstart"] = int(line[8:16])
            header["nxend"] = int(line[16:24])
            header["ny"] = int(line[24:32])
            header["nystart"] = int(line[32:40])
            header["nyend"] = int(line[40:48])
            header["nz"] = int(line[48:56])
            header["nzstart"] = int(line[56:64])
            header["nzend"] = int(line[64:72])

            line = volume.readline()
            header["xlength"] = float(line[0:12])
            header["ylength"] = float(line[12:24])
            header["zlength"] = float(line[24:36])
            header["alpha"] = float(line[36:48])
            header["beta"] = float(line[48:60])
            header["gamma"] = float(line[60:72])

            header["order"] = volume.readline()[0:3]

            self.header = header

    @property
    def voxelspacing(self):
        return self.header["xlength"] / float(self.header["nx"])

    @property
    def origin(self):
        return [self.voxelspacing * x for x in [self.header["nxstart"], self.header["nystart"], self.header["nzstart"]]]

    @property
    def density(self):
        with open(self.source) as volumefile:
            for _ in range(2 + len(self.header["label"]) + 3):
                volumefile.readline()
            nx = self.header["nx"]
            ny = self.header["ny"]
            nz = self.header["nz"]

            array = np.zeros((nz, ny, nx), dtype=np.float64)

            xextend = self.header["nxend"] - self.header["nxstart"] + 1
            yextend = self.header["nyend"] - self.header["nystart"] + 1
            zextend = self.header["nzend"] - self.header["nzstart"] + 1

            nslicelines = int(np.ceil(xextend * yextend / 6.0))
            for i in range(zextend):
                values = []
                for _ in range(nslicelines):
                    line = volumefile.readline()
                    for n in range(len(line) // 12):
                        value = float(line[n * 12 : (n + 1) * 12])
                        values.append(value)
                array[i, :yextend, :xextend] = np.float64(values).reshape(yextend, xextend)

        return array


def to_xplor(outfile: str | Path, volume: Volume, label: list = []):  # noqa: B006
    nz, ny, nx = volume.shape
    xstart, ystart, zstart = [int(round(x)) for x in volume.start]
    xlength, ylength, zlength = volume.dimensions
    alpha = beta = gamma = 90.0

    nlabel = len(label)
    with open(outfile, "w") as out:
        out.write("\n")
        out.write(f"{nlabel + 1:>8d} !NTITLE\n")
        # CNS requires at least one REMARK line
        out.write("REMARK\n")
        for n in range(nlabel):
            out.write("".join(["REMARK ", label[n], "\n"]))

        out.write(
            ("{:>8d}" * 9 + "\n").format(
                nx, xstart, xstart + nx - 1, ny, ystart, ystart + ny - 1, nz, zstart, zstart + nz - 1
            )
        )
        out.write(("{:12.5E}" * 6 + "\n").format(xlength, ylength, zlength, alpha, beta, gamma))
        out.write("ZYX\n")
        # FIXME very inefficient way of writing out the volume ...
        for z in range(nz):
            out.write(f"{z:>8d}\n")
            n = 0
            for y in range(ny):
                for x in range(nx):
                    out.write("%12.5E" % volume.array[z, y, x])  # noqa: UP031
                    n += 1
                    if (n % 6) == 0:
                        out.write("\n")
            if (nx * ny) % 6 > 0:
                out.write("\n")
        out.write(f"{-9999:>8d}\n")
        out.write(f"{volume.array.mean():12.4E} {volume.array.std():12.4E} ")



import gzip
import operator
from collections import OrderedDict, defaultdict
from collections.abc import Sequence
from io import BufferedReader, TextIOWrapper
from string import capwords

import numpy as np

from .elements import ELEMENTS

# records
MODEL = "MODEL "
ATOM = "ATOM  "
HETATM = "HETATM"
TER = "TER   "

MODEL_LINE = "MODEL " + " " * 4 + "{:>4d}\n"
ENDMDL_LINE = "ENDMDL\n"
TER_LINE = "TER   " + "{:>5d}" + " " * 6 + "{:3s}" + " " + "{:1s}" + "{:>4d}" + "{:1s}" + " " * 53 + "\n"
ATOM_LINE = (
    "{:6s}"
    + "{:>5d}"
    + " "
    + "{:4s}"
    + "{:1s}"
    + "{:3s}"
    + " "
    + "{:1s}"
    + "{:>4d}"
    + "{:1s}"
    + " " * 3
    + "{:8.3f}" * 3
    + "{:6.2f}" * 2
    + " " * 10
    + "{:<2s}"
    + "{:2s}\n"
)
END_LINE = "END   \n"

ATOM_DATA = ["record", "id", "name", "alt", "resn", "chain", "resi", "i", "x", "y", "z", "q", "b", "e", "charge"]
TER_DATA = ["id", "resn", "chain", "resi", "i"]


def parse_pdb(infile):
    f = _open_maybe_gzipped_text_file(infile)

    pdb = defaultdict(list)
    model_number = 1
    for line in f:
        record = line[:6]
        if record in (ATOM, HETATM):
            pdb["model"].append(model_number)
            pdb["record"].append(record)
            pdb["id"].append(int(line[6:11]))
            name = line[12:16].strip()
            pdb["name"].append(name)
            pdb["alt"].append(line[16])
            pdb["resn"].append(line[17:20].strip())
            pdb["chain"].append(line[21])
            pdb["resi"].append(int(line[22:26]))
            pdb["i"].append(line[26])
            pdb["x"].append(float(line[30:38]))
            pdb["y"].append(float(line[38:46]))
            pdb["z"].append(float(line[46:54]))
            pdb["q"].append(float(line[54:60]))
            pdb["b"].append(float(line[60:66]))
            # Be forgiving when determining the element
            e = line[76:78].strip()
            if not e:
                # If element is not given, take the first non-numeric letter of
                # the name as element.
                for e in name:
                    if e.isalpha():
                        break
            pdb["e"].append(e)
            pdb["charge"].append(line[78:80].strip())
        elif record == MODEL:
            model_number = int(line[10:14])
    f.close()
    return pdb


def tofile(pdb, out):
    f = open(out, "w")  # noqa: SIM115

    nmodels = len(set(pdb["model"]))
    natoms = len(pdb["id"])
    natoms_per_model = natoms // nmodels

    for nmodel in range(nmodels):
        offset = int(nmodel * natoms_per_model)
        # write MODEL record
        if nmodels > 1:
            f.write(MODEL_LINE.format(nmodel + 1))
        prev_chain = pdb["chain"][offset]
        for natom in range(natoms_per_model):
            index = int(offset + natom)

            # write TER record
            current_chain = pdb["chain"][index]
            if prev_chain != current_chain:
                prev_record = pdb["record"][index - 1]
                if prev_record == ATOM:
                    line_data = [pdb[data][index - 1] for data in TER_DATA]
                    line_data[0] += 1
                    f.write(TER_LINE.format(*line_data))
                prev_chain = current_chain

            # write ATOM/HETATM record
            line_data = [pdb[data][index] for data in ATOM_DATA]
            # take care of the rules for atom name position
            e = pdb["e"][index]
            name = pdb["name"][index]
            if len(e) == 1 and len(name) != 4:
                line_data[2] = " " + name
            f.write(ATOM_LINE.format(*line_data))

        # write ENDMDL record
        if nmodels > 1:
            f.write(ENDMDL_LINE)

    f.write(END_LINE)
    f.close()


def pdb_dict_to_array(pdb):
    dtype = [
        ("record", np.str_, 6),
        ("id", np.int32),
        ("name", np.str_, 4),
        ("alt", np.str_, 1),
        ("resn", np.str_, 4),
        ("chain", np.str_, 2),
        ("resi", np.int32),
        ("i", np.str_, 1),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
        ("q", np.float64),
        ("b", np.float64),
        ("e", np.str_, 2),
        ("charge", np.str_, 2),
        ("model", np.int32),
    ]

    natoms = len(pdb["id"])
    pdb_array = np.empty(natoms, dtype=dtype)
    for data in ATOM_DATA:
        pdb_array[data] = pdb[data]
    pdb_array["model"] = pdb["model"]
    return pdb_array


def pdb_array_to_dict(pdb_array):
    pdb = defaultdict(list)
    for data in ATOM_DATA:
        pdb[data] = pdb_array[data].tolist()
    pdb["model"] = pdb_array["model"].tolist()
    return pdb


class Structure:
    @classmethod
    def fromfile(cls, fid):
        """Initialize Structure from PDB-file"""
        try:
            fname = fid.name
        except AttributeError:
            fname = fid

        if fname.endswith(".pdb") or fname.endswith(".ent") or fname.endswith(".pdb.gz") or fname.endswith(".ent.gz"):
            arr = pdb_dict_to_array(parse_pdb(fid))
        elif fname.endswith(".cif") or fname.endswith(".cif.gz"):
            arr = mmcif_dict_to_array(parse_mmcif(fid))
        else:
            raise OSError("Filetype not recognized.")
        return cls(arr)

    def __init__(self, pdb):
        self.data = pdb

    @property
    def atomnumber(self):
        """Return array of atom numbers"""
        return self._get_property("number")

    @property
    def chain_list(self):
        return np.unique(self.data["chain"])

    def combine(self, structure):
        return Structure(np.hstack((self.data, structure.data)))

    @property
    def coor(self):
        """Return the coordinates"""
        return np.asarray([self.data["x"], self.data["y"], self.data["z"]])

    def duplicate(self):
        """Duplicate the object"""
        return Structure(self.data.copy())

    def _get_property(self, ptype):
        elements, ind = np.unique(self.data["e"], return_inverse=True)
        return np.asarray([getattr(ELEMENTS[capwords(e)], ptype) for e in elements], dtype=np.float64)[ind]

    @property
    def mass(self):
        return self._get_property("mass")

    def rmsd(self, structure):
        return np.sqrt(((self.coor - structure.coor) ** 2).mean() * 3)

    def rotate(self, rotmat):
        """Rotate atoms"""
        self.data["x"], self.data["y"], self.data["z"] = np.asmatrix(rotmat) * np.asmatrix(self.coor)

    def select(self, identifier, values, loperator="==", return_ind=False) -> "Structure":
        """A simple way of selecting atoms"""
        if loperator == "==":
            oper = operator.eq
        elif loperator == "<":
            oper = operator.lt
        elif loperator == ">":
            oper = operator.gt
        elif loperator == ">=":
            oper = operator.ge
        elif loperator == "<=":
            oper = operator.le
        elif loperator == "!=":
            oper = operator.ne
        else:
            raise ValueError("Logic operator not recognized.")

        if not isinstance(values, Sequence) or isinstance(values, str):
            values = (values,)

        selection = oper(self.data[identifier], values[0])
        if len(values) > 1:
            for v in values[1:]:
                if loperator == "!=":
                    selection &= oper(self.data[identifier], v)
                else:
                    selection |= oper(self.data[identifier], v)

        if return_ind:
            return selection
        else:
            return Structure(self.data[selection])

    @property
    def sequence(self):
        resids, indices = np.unique(self.data["resi"], return_index=True)
        return self.data["resn"][indices]

    def translate(self, trans):
        """Translate atoms"""
        self.data["x"] += trans[0]
        self.data["y"] += trans[1]
        self.data["z"] += trans[2]

    def tofile(self, fid):
        """Write instance to PDB-file"""
        tofile(pdb_array_to_dict(self.data), fid)

    @property
    def rvdw(self):
        return self._get_property("vdwrad")


def _open_maybe_gzipped_text_file(infile: str | BufferedReader | TextIOWrapper) -> TextIOWrapper:
    if isinstance(infile, BufferedReader):
        if infile.name.endswith(".gz"):
            return gzip.open(infile, "rt")
        else:
            return TextIOWrapper(infile, encoding="utf-8")
    elif isinstance(infile, str):
        if infile.endswith(".gz"):
            return gzip.open(infile, "rt")
        else:
            return open(infile)
    elif not isinstance(infile, TextIOWrapper):
        raise TypeError("Input should either be a file object or string.")
    return infile


def parse_mmcif(infile):
    infile = _open_maybe_gzipped_text_file(infile)

    atom_site = OrderedDict()
    with infile as f:
        for line in f:
            if line.startswith("_atom_site."):
                words = line.split(".")
                atom_site[words[1].strip()] = []

            if line.startswith("ATOM"):
                words = line.split()
                for key, word in zip(atom_site, words, strict=False):
                    atom_site[key].append(word)
    infile.close()
    return atom_site


def mmcif_dict_to_array(atom_site):
    natoms = len(atom_site["id"])
    dtype = [
        ("record", np.str_, 6),
        ("id", np.int32),
        ("name", np.str_, 4),
        ("alt", np.str_, 1),
        ("resn", np.str_, 4),
        ("chain", np.str_, 2),
        ("resi", np.int32),
        ("i", np.str_, 1),
        ("x", np.float64),
        ("y", np.float64),
        ("z", np.float64),
        ("q", np.float64),
        ("b", np.float64),
        ("e", np.str_, 2),
        ("charge", np.str_, 2),
        ("model", np.int32),
    ]

    cifdata = np.zeros(natoms, dtype=dtype)
    cifdata["record"] = "ATOM  "
    cifdata["id"] = atom_site["id"]
    cifdata["name"] = atom_site["label_atom_id"]
    cifdata["resn"] = atom_site["label_comp_id"]
    cifdata["chain"] = atom_site["label_asym_id"]
    cifdata["resi"] = atom_site["label_seq_id"]
    cifdata["x"] = atom_site["Cartn_x"]
    cifdata["y"] = atom_site["Cartn_y"]
    cifdata["z"] = atom_site["Cartn_z"]
    cifdata["q"] = atom_site["occupancy"]
    cifdata["b"] = atom_site["B_iso_or_equiv"]
    cifdata["e"] = atom_site["type_symbol"]
    cifdata["charge"] = atom_site["pdbx_formal_charge"]
    cifdata["model"] = atom_site["pdbx_PDB_model_num"]
    return cifdata



#! ../env/bin/python


import logging
import warnings
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, BooleanOptionalAction, FileType
from functools import partial
from os.path import abspath, join, splitext
from pathlib import Path
from time import time
from typing import TYPE_CHECKING, BinaryIO

import numpy as np
from rich.logging import RichHandler
from tqdm import TqdmExperimentalWarning
from tqdm.auto import tqdm
from tqdm.rich import tqdm as rich_tqdm

from powerfit_em import (
    Structure,
    Volume,
    __version__,
    determine_core_indices,
    proportional_orientations,
    quat_to_rotmat,
    structure_to_shape_like,
)
from powerfit_em.analyzer import Analyzer
from powerfit_em.helpers import fisher_sigma, opencl_available, write_fits_to_pdb
from powerfit_em.powerfitter import PowerFitter
from powerfit_em.report import generate_report
from powerfit_em.volume import extend, nearest_multiple2357, resample, trim

if TYPE_CHECKING:
    import pyopencl as cl  # noqa: I001


logger = logging.getLogger(__name__)

warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)


def make_parser():
    """Create the command-line argument parser.
    Any changes here should be reflected in the documentation at docs/manual.md#parameters"""
    p = ArgumentParser(prog="powerfit", formatter_class=ArgumentDefaultsHelpFormatter)

    p.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    # Positional arguments
    p.add_argument(
        "target",
        type=FileType("rb"),
        help="Target density map to fit the model in. Data should either be in CCP4 or MRC format",
    )
    p.add_argument("resolution", type=float, help="Resolution of map in angstrom")
    p.add_argument(
        "template",
        type=FileType("rb"),
        help="Atomic model to be fitted in the density. "
        "Format should either be PDB or mmCIF. "
        "Valid extensions are .pdb, .ent, .cif, .pdb.gz, .ent.gz, .cif.gz",
    )

    # Optional arguments and flags
    p.add_argument(
        "-a",
        "--angle",
        dest="angle",
        type=float,
        default=10,
        metavar="<float>",
        help="Rotational sampling density in degree. Increasing "
        "this number by a factor of 2 results in approximately "
        "8 times more rotations sampled.",
    )
    # Scoring flags
    p.add_argument(
        "-nl",
        "--no-laplace",
        dest="no_laplace",
        action="store_true",
        help="Do not use the Laplace pre-filter density data.",
    )
    p.add_argument(
        "-ncw",
        "--no-core-weighted",
        dest="no_core_weighted",
        action="store_true",
        help="Do not use core-weighted local cross-correlation score.",
    )
    # Resampling
    p.add_argument(
        "-nr",
        "--no-resampling",
        dest="no_resampling",
        action="store_true",
        help="Do not resample the density map.",
    )
    p.add_argument(
        "-rr",
        "--resampling-rate",
        dest="resampling_rate",
        type=float,
        default=2,
        metavar="<float>",
        help="Resampling rate compared to Nyquist.",
    )
    # Trimming related
    p.add_argument(
        "-nt",
        "--no-trimming",
        dest="no_trimming",
        action="store_true",
        help="Do not trim the density map.",
    )
    p.add_argument(
        "-tc",
        "--trimming-cutoff",
        dest="trimming_cutoff",
        type=float,
        default=None,
        metavar="<float>",
        help="Intensity cutoff to which the map will be trimmed. Default is 10 percent of maximum intensity.",
    )
    # Selection parameter
    p.add_argument(
        "-c",
        "--chain",
        dest="chain",
        type=str,
        default=None,
        metavar="<char>",
        help=(
            "The chain IDs of the structure to be fitted. "
            "Multiple chains can be selected using a comma separated list, i.e. -c A,B,C. "
            "Default is the whole structure."
        ),
    )
    # Output parameters
    p.add_argument(
        "-d",
        "--directory",
        dest="directory",
        type=abspath,
        default=".",
        metavar="<dir>",
        help="Directory where the results are stored.",
    )
    p.add_argument(
        "-n",
        "--num",
        dest="num",
        type=int,
        default=10,
        metavar="<int>",
        help="Number of models written to file. This number will be capped if less solutions are found as requested.",
    )
    # Computational resources parameters
    p.add_argument(
        "-g",
        "--gpu",
        dest="gpu",
        nargs="?",
        const="0:0",
        default=None,
        metavar="[<platform>:<device>]",
        help="Off-load the intensive calculations to the GPU. Optionally specify platform and device as <platform>:<device> (e.g., --gpu 0:3). If not specified, uses first device in first platform. If omitted, does not use GPU.",  # noqa: E501
    )
    p.add_argument(
        "-p",
        "--nproc",
        dest="nproc",
        type=int,
        default=1,
        metavar="<int>",
        help="Number of processors used during search. "
        "The number will be capped at the total number "
        "of available processors on your machine.",
    )
    p.add_argument(
        "--log-level",
        dest="log_level",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level.",
    )
    p.add_argument(
        "--delimiter",
        dest="delimiter",
        type=str,
        default=None,
        metavar="<str>",
        help="Delimiter used in the 'solutions.out' file. For example use ',' or '\\t'. Defaults to fixed width.",
    )
    p.add_argument(
        "--progressbar",
        dest="progressbar",
        action=BooleanOptionalAction,
        default=True,
        help="Show a progress bar during the search. Disabling the progressbar will improve performance.",
    )
    p.add_argument(
        "--report",
        dest="report",
        action="store_true",
        help="Generate a html report with Mol* 3D viewer of the fitted models.",
    )

    return p


def parse_args():
    """Parse command-line options."""
    p = make_parser()
    args = p.parse_args()

    return args


def get_filetype_template(fname):
    """Determine the file type from the file extension."""
    ext = splitext(fname)[1][1:]
    if ext in ["pdb", "ent"]:
        ft = "pdb"
    elif ext in ["map", "ccp4"]:
        ft = "map"
    else:
        msg = f"Filetype of file {fname} is not recognized."
        raise OSError(msg)
    return ft


def configure_logging(log_file, log_level="INFO"):
    for handler in logging.root.handlers:
        logging.root.removeHandler(handler)

    # Write log messages to a file
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(log_level)
        file_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
        logging.root.addHandler(file_handler)

    # Write to console with rich formatting
    console_handler = RichHandler(show_time=False, show_path=False, show_level=False)
    console_handler.setLevel(log_level)
    logging.root.addHandler(console_handler)
    logger.setLevel(log_level)


def main():
    args = parse_args()

    Path(args.directory).mkdir(exist_ok=True)
    configure_logging(join(args.directory, "powerfit.log"), args.log_level)

    progress = partial(rich_tqdm, desc="Processing rotations", unit="rot") if args.progressbar else None

    powerfit(
        target_volume=args.target,
        resolution=args.resolution,
        template_structure=args.template,
        angle=args.angle,
        laplace=not args.no_laplace,
        core_weighted=not args.no_core_weighted,
        no_resampling=args.no_resampling,
        resampling_rate=args.resampling_rate,
        no_trimming=args.no_trimming,
        trimming_cutoff=args.trimming_cutoff,
        chain=args.chain,
        directory=args.directory,
        num=args.num,
        gpu=args.gpu,
        nproc=args.nproc,
        delimiter=args.delimiter,
        progress=progress,
    )
    if args.report:
        # Report shows all options that affect the fitting
        options = {
            "resolution": args.resolution,
            "angle": args.angle,
            "no laplace": args.no_laplace,
            "no core weighted": args.no_core_weighted,
            "no resampling": args.no_resampling,
            "resampling rate": args.resampling_rate,
            "no trimming": args.no_trimming,
            "trimming cutoff": args.trimming_cutoff,
        }
        generate_report(args.directory, args.target.name, args.num, args.delimiter, options=options)


def get_gpu_queue(gpu: str) -> "cl.CommandQueue":
    """Request an OpenCL Queue."""
    if not opencl_available():
        msg = "Running on GPU requires the pyopencl package, however importing pyopencl failed."
        raise ValueError(msg)
    else:
        import pyopencl as cl
    # TODO allow to omit platform, so gpu='4' runs 5th device on first platform
    if ":" in gpu:
        platform_idx, device_idx = map(int, gpu.split(":"))
    else:
        platform_idx, device_idx = 0, 0
    platforms = cl.get_platforms()
    if platform_idx >= len(platforms):
        raise RuntimeError(f"Requested OpenCL platform {platform_idx} not found.")
    platform = platforms[platform_idx]
    devices = platform.get_devices()
    if device_idx >= len(devices):
        raise RuntimeError(f"Requested OpenCL device {device_idx} not found on platform {platform_idx}.")
    context = cl.Context(devices=[devices[device_idx]])
    return cl.CommandQueue(context, device=devices[device_idx])


def setup_target(
    target_volume: BinaryIO,
    resolution: float,
    no_resampling: bool,
    resampling_rate: float,
    no_trimming: bool,
    trimming_cutoff: float | None,
) -> Volume:
    """Load and preprocess the target density."""
    logger.info(f"Target file read from: {abspath(target_volume.name):s}")
    target = Volume.fromfile(target_volume)
    logger.info(f"Target resolution: {resolution:.2f}")
    logger.info(("Initial shape of density:" + " {:d}" * 3).format(*target.shape))

    # Resample target density if requested
    if not no_resampling:
        factor = 2 * resampling_rate * target.voxelspacing / resolution
        if factor < 0.9:
            target = resample(target, factor)
            logger.info(("Shape after resampling:" + " {:d}" * 3).format(*target.shape))

    # Trim target density if requested
    if not no_trimming:
        if trimming_cutoff is None:
            trimming_cutoff = target.array.max() / 10
        target = trim(target, trimming_cutoff)
        logger.info(("Shape after trimming:" + " {:d}" * 3).format(*target.shape))

    # Extend the density to a multiple of 2, 3, 5, and 7 for clFFT
    extended_shape = [nearest_multiple2357(n) for n in target.shape]
    target = extend(target, extended_shape)
    logger.info(("Shape after extending:" + " {:d}" * 3).format(*target.shape))
    return target


def setup_template_structure(
    template_structure: BinaryIO, chain: str | None, target: Volume, resolution: float, core_weighted: bool
) -> tuple[Structure, Volume, Volume, float]:
    """Load structure, setup template and mask, precompute Fisher sigma for scoring."""
    # Read in structure or high-resolution map
    logger.info(f"Template file read from: {abspath(template_structure.name):s}")
    structure = Structure.fromfile(template_structure)
    if chain is not None:
        logger.info("Selecting chains: " + chain)
        structure = structure.select("chain", chain.split(","))
    if structure.data.size == 0:
        raise ValueError("No atoms were selected.")

    # Move structure to origin of density
    structure.translate(target.origin - structure.coor.mean(axis=1))
    template = structure_to_shape_like(
        target,
        structure.coor,
        resolution=resolution,
        weights=structure.atomnumber,
        shape="vol",
    )
    mask = structure_to_shape_like(target, structure.coor, resolution=resolution, shape="mask")

    # Apply core-weighted mask if requested
    if core_weighted:
        logger.info("Calculating core-weighted mask.")
        mask.array = determine_core_indices(mask.array)

    # calculate the molecular volume of the structure
    mv = (
        structure_to_shape_like(
            target,
            structure.coor,
            resolution=resolution,
            radii=structure.rvdw,
            shape="mask",
        ).array.sum()
        * target.voxelspacing**3
    )
    z_sigma = fisher_sigma(mv, resolution)
    return structure, template, mask, z_sigma


def setup_rotational_matrix(angle: float) -> np.ndarray:
    logger.info("Reading in rotations.")
    q, _, degree = proportional_orientations(angle)
    rotmat = quat_to_rotmat(q)
    logger.info(f"Requested rotational sampling density: {angle:.2f}")
    logger.info(f"Real rotational sampling density: {degree}")
    return rotmat


def powerfit(
    target_volume: BinaryIO,
    resolution: float,
    template_structure: BinaryIO,
    angle: float = 10,
    laplace: bool = False,
    core_weighted: bool = False,
    no_resampling: bool = False,
    resampling_rate: float = 2,
    no_trimming: bool = False,
    trimming_cutoff: float | None = None,
    chain: str | None = None,
    directory: str = ".",
    num: int = 10,
    gpu: str | None = None,
    nproc: int = 1,
    delimiter: str | None = None,
    progress: partial[tqdm] | None = tqdm,
):
    time0 = time()
    Path(directory).mkdir(exist_ok=True)

    # Get GPU queue if requested
    queue = None
    if gpu:
        queue = get_gpu_queue(gpu)

    target = setup_target(target_volume, resolution, no_resampling, resampling_rate, no_trimming, trimming_cutoff)
    structure, template, mask, z_sigma = setup_template_structure(
        template_structure, chain, target, resolution, core_weighted
    )
    rotmat = setup_rotational_matrix(angle)

    pf = PowerFitter(target, rotmat, template, mask, queue, nproc, laplace=laplace)
    if gpu:
        logger.info("Using GPU-accelerated search.")
    else:
        logger.info(f"Requested number of processors: {nproc:d}")

    logger.info("Starting search")

    time1 = time()
    pf.scan(progress=progress)
    dtime = time() - time1
    if dtime < 10:
        logger.info(f"Time for search: {dtime:.3f} s")
    else:
        logger.info("Time for search: {:.0f}m {:.0f}s".format(*divmod(dtime, 60)))
    logger.info("Analyzing results")

    analyzer = Analyzer(
        pf.lcc,
        rotmat,
        pf.rot,
        voxelspacing=target.voxelspacing,
        origin=target.origin,
        z_sigma=z_sigma,
    )

    logger.info("Writing solutions to file.")
    Volume(pf.lcc, target.voxelspacing, target.origin).tofile(join(directory, "lcc.mrc"))
    analyzer.tofile(join(directory, "solutions.out"), delimiter=delimiter)

    logger.info("Writing PDBs to file.")
    n = min(num, len(analyzer.solutions))
    write_fits_to_pdb(structure, analyzer.solutions[:n], basename=join(directory, "fit"))

    logger.info("Total time: {:.0f}m {:.0f}s".format(*divmod(time() - time0, 60)))


def powerfit_many(
    target_volume: Path,
    resolution: float,
    template_structures: list[Path],
    angle: float = 10,
    laplace: bool = False,
    core_weighted: bool = False,
    no_resampling: bool = False,
    resampling_rate: float = 2,
    no_trimming: bool = False,
    trimming_cutoff: float | None = None,
    gpu: str | None = None,
    reuse: bool = True,
    nproc: int = 1,
) -> list[list[list[float]]]:
    """Run powerfit on multiple templates, returning the solution table for each.

    For a slight efficiency boost, and to avoid continuously creating many new OpenCL
    queues, the queues are reused. This can be disabled by setting reuse=False
    Outer list is same order as template_structures. Middle list is ordered on cc score. Inner list is the data for a solution.  Each solution has the following columns: cc Fish-z rel-z x y z a11 a12 a13 a21 a22 a23 a31 a32 a33.
    """  # noqa: E501
    time0 = time()

    # Get GPU queue if requested
    queue = None
    if gpu:
        queue = get_gpu_queue(gpu)

    with target_volume.open("rb") as f:
        target = setup_target(
            f,
            resolution,
            no_resampling,
            resampling_rate,
            no_trimming,
            trimming_cutoff,
        )

    template_vars: list[tuple[Structure, Volume, Volume, float]] = []
    for template_structure in template_structures:
        with template_structure.open("r") as f:
            template_vars.append(setup_template_structure(f, None, target, resolution, core_weighted))
    rotmat = setup_rotational_matrix(angle)

    if gpu:
        logger.info("Using GPU-accelerated search.")
    else:
        logger.info(f"Requested number of processors: {nproc}.")

    logger.info(f"Starting search, analysing {len(template_structures)} structures.")

    time1 = time()
    results: list[tuple[np.ndarray, np.ndarray]] = []
    pf: PowerFitter | None = None
    for i in range(len(template_vars)):
        _, template, mask, z_sigma = template_vars[i]
        if pf is None or not reuse or not gpu and nproc > 1:
            pf = PowerFitter(target, rotmat, template, mask, queue, nproc, laplace=laplace)
        else:
            pf.set_template(template, mask)

        pf.scan(progress=None)
        results.append((pf.lcc, pf.rot))

    dtime = time() - time1
    if dtime < 10:
        logger.info(f"Time for searches: {dtime:.3f} s")
    else:
        logger.info("Time for searches: {:.0f}m {:.0f}s".format(*divmod(dtime, 60)))

    logger.info("Analyzing results")

    analysis_results: list[Analyzer] = []
    for result, template_var, _ in zip(results, template_vars, template_structures, strict=True):
        lcc, rot = result
        _, _, _, z_sigma = template_var

        analysis = Analyzer(
            lcc,
            rotmat,
            rot,
            voxelspacing=target.voxelspacing,
            origin=target.origin,
            z_sigma=z_sigma,
        )
        analysis_results.append(analysis)

    logger.info("Total time: {:.0f}m {:.0f}s".format(*divmod(time() - time0, 60)))
    return [r.solutions for r in analysis_results]


if __name__ == "__main__":
    main()



import unittest

import numpy as np

from powerfit_em.rotations import euler, euler_to_rotmat, random_rotmat


class TestRotations(unittest.TestCase):
    def test_random_rotmat(self):
        rotmat = random_rotmat()
        # the determinant of a rotmat is 1
        self.assertAlmostEqual(np.linalg.det(rotmat), 1)

    def test_euler(self):
        # 90 z-rotation
        angle = np.radians(90)
        out = euler(angle, "z")
        answer = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
        self.assertTrue(np.allclose(answer, out))

        # 90 y-rotation
        out = euler(angle, "y")
        answer = [[0, 0, 1], [0, 1, 0], [-1, 0, 0]]
        self.assertTrue(np.allclose(answer, out))

        # 90 x-rotation
        out = euler(angle, "x")
        answer = [[1, 0, 0], [0, 0, -1], [0, 1, 0]]
        self.assertTrue(np.allclose(answer, out))

    def test_euler_to_rotmat(self):
        # 90 z-rotation
        angles = [np.radians(x) for x in (0, 0, 90)]
        answer = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]
        out = euler_to_rotmat(angles)
        self.assertTrue(np.allclose(answer, out))


if __name__ == "__main__":
    unittest.main()
